@@ -82,13 +82,14 @@ module Sequel
       # columns are deleted.  This method should generally not be called
       # directly by user code.
       def clone(opts = (return self; nil))
+        # return self used above because clone is called by almost all
+        # other query methods, and it is the fastest approach
         c = super(:freeze=>false)
         c.opts.merge!(opts)
         unless opts.each_key{|o| break if COLUMN_CHANGE_OPTS.include?(o)}
           c.clear_columns_cache
         end
         c.freeze
-        c
       end
     else
       # :nocov:
@@ -114,10 +115,19 @@ module Sequel
     #  DB[:items].distinct # SQL: SELECT DISTINCT * FROM items
     #  DB[:items].order(:id).distinct(:id) # SQL: SELECT DISTINCT ON (id) * FROM items ORDER BY id
     #  DB[:items].order(:id).distinct{func(:id)} # SQL: SELECT DISTINCT ON (func(id)) * FROM items ORDER BY id
+    #
+    # There is support for emualting the DISTINCT ON support in MySQL, but it
+    # does not support the ORDER of the dataset, and also doesn't work in many
+    # cases if the ONLY_FULL_GROUP_BY sql_mode is used, which is the default on
+    # MySQL 5.7.5+.
     def distinct(*args, &block)
       virtual_row_columns(args, block)
-      raise(InvalidOperation, "DISTINCT ON not supported") if !args.empty? && !supports_distinct_on?
-      clone(:distinct => args.freeze)
+      if args.empty?
+        cached_dataset(:_distinct_ds){clone(:distinct => EMPTY_ARRAY)}
+      else
+        raise(InvalidOperation, "DISTINCT ON not supported") unless supports_distinct_on?
+        clone(:distinct => args.freeze)
+      end
     end
 
     # Adds an EXCEPT clause using a second dataset object.
@@ -190,7 +200,6 @@ module Sequel
         c = _clone(:freeze=>false)
         c.send(:_extension!, a)
         c.freeze
-        c
       end
     else
       # :nocov:
@@ -274,11 +283,15 @@ module Sequel
     def from_self(opts=OPTS)
       fs = {}
       @opts.keys.each{|k| fs[k] = nil unless non_sql_option?(k)}
-      c = clone(fs).from(opts[:alias] ? as(opts[:alias], opts[:column_aliases]) : self)
-      if cols = _columns
-        c.send(:columns=, cols)
+      pr = proc do
+        c = clone(fs).from(opts[:alias] ? as(opts[:alias], opts[:column_aliases]) : self)
+        if cols = _columns
+          c.send(:columns=, cols)
+        end
+        c
       end
-      c
+
+      cache ? cached_dataset(:_from_self_ds, &pr) : pr.call
     end
 
     # Match any of the columns to any of the patterns. The terms can be
@@ -616,7 +629,7 @@ module Sequel
     #   DB.from(:a, DB[:b].where(Sequel[:a][:c]=>Sequel[:b][:d]).lateral)
     #   # SELECT * FROM a, LATERAL (SELECT * FROM b WHERE (a.c = b.d))
     def lateral
-      clone(:lateral=>true)
+      cached_dataset(:_lateral_ds){clone(:lateral=>true)}
     end
 
     # If given an integer, the dataset will contain only the first l results.
@@ -672,6 +685,18 @@ module Sequel
       cached_dataset(:_naked_ds){with_row_proc(nil)}
     end
 
+    # Returns a copy of the dataset that will raise a DatabaseLockTimeout instead
+    # of waiting for rows that are locked by another transaction
+    #
+    #   DB[:items].for_update.nowait
+    #   # SELECT * FROM items FOR UPDATE NOWAIT
+    def nowait
+      cached_dataset(:_nowait_ds) do
+        raise(Error, 'This dataset does not support raises errors instead of waiting for locked rows') unless supports_nowait?
+        clone(:nowait=>true)
+      end
+    end
+
     # Returns a copy of the dataset with a specified order. Can be safely combined with limit.
     # If you call limit with an offset, it will override override the offset if you've called
     # offset first.
@@ -755,15 +780,20 @@ module Sequel
     #
     #   DB[:items].where(id: 1).qualify(:i)
     #   # SELECT i.* FROM items WHERE (i.id = 1)
-    def qualify(table=first_source)
+    def qualify(table=(cache=true; first_source))
       o = @opts
       return self if o[:sql]
-      h = {}
-      (o.keys & QUALIFY_KEYS).each do |k|
-        h[k] = qualified_expression(o[k], table)
+
+      pr = proc do
+        h = {}
+        (o.keys & QUALIFY_KEYS).each do |k|
+          h[k] = qualified_expression(o[k], table)
+        end
+        h[:select] = [SQL::ColumnAll.new(table)].freeze if !o[:select] || o[:select].empty?
+        clone(h)
       end
-      h[:select] = [SQL::ColumnAll.new(table)].freeze if !o[:select] || o[:select].empty?
-      clone(h)
+
+      cache ? cached_dataset(:_qualify_ds, &pr) : pr.call
     end
 
     # Modify the RETURNING clause, only supported on a few databases.  If returning
@@ -785,8 +815,15 @@ module Sequel
     #     # hash for each row deleted, with values for all columns
     #   end
     def returning(*values)
-      raise Error, "RETURNING is not supported on #{db.database_type}" unless supports_returning?(:insert)
-      clone(:returning=>values.freeze)
+      if values.empty?
+        cached_dataset(:_returning_ds) do
+          raise Error, "RETURNING is not supported on #{db.database_type}" unless supports_returning?(:insert)
+          clone(:returning=>EMPTY_ARRAY)
+        end
+      else
+        raise Error, "RETURNING is not supported on #{db.database_type}" unless supports_returning?(:insert)
+        clone(:returning=>values.freeze)
+      end
     end
 
     # Returns a copy of the dataset with the order reversed. If no order is
@@ -831,7 +868,7 @@ module Sequel
     #   DB[:items].select_all(:items, :foo) # SELECT items.*, foo.* FROM items
     def select_all(*tables)
       if tables.empty?
-        clone(:select => nil)
+        cached_dataset(:_select_all_ds){clone(:select => nil)}
       else
         select(*tables.map{|t| i, a = split_alias(t); a || i}.map!{|t| SQL::ColumnAll.new(t)}.freeze)
       end
@@ -1009,6 +1046,16 @@ module Sequel
       add_filter(:where, cond, &block)
     end
     
+    # Return a clone of the dataset with an addition named window that can be
+    # referenced in window functions. See Sequel::SQL::Window for a list of
+    # options that can be passed in. Example:
+    #
+    #   DB[:items].window(:w, :partition=>:c1, :order=>:c2)
+    #   # SELECT * FROM items WINDOW w AS (PARTITION BY c1 ORDER BY c2)
+    def window(name, opts)
+      clone(:window=>((@opts[:window]||EMPTY_ARRAY) + [[name, SQL::Window.new(opts)].freeze]).freeze)
+    end
+
     # Add a common table expression (CTE) with the given name and a dataset that defines the CTE.
     # A common table expression acts as an inline view for the query.
     # Options:
@@ -1067,7 +1114,6 @@ module Sequel
         c.extend(*mods) unless mods.empty?
         c.extend(DatasetModule.new(&block)) if block
         c.freeze
-        c
       end
     else
       # :nocov:
